/*===============================================================================
Title: dwLinkItWifi.ino
Description: deviceWISE LinkIt with WiFi Sample Program

Copyright (c)2014, 2015 ILS Technology.
http://www.ilstechnology.com
http://www.devicewise.com


Legal Disclaimer:

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

    Neither the name of ILS Technology nor Telit nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===================================================================================*/


  #include <SPI.h>
  #include <LWiFi.h>
  #include <LWiFiClient.h>
  #include <DHT.h>                          //Grove Base Include
  #include <Wire.h>  
  #include <Suli.h>  
  #include <Seeed_LED_Bar_Arduino.h>        //LED Bar
  
  #include <dwOpenClientLinkIt.h>

  
  //Wi-Fi Network - SSID / Password
  #define WIFI_SSID          "UserGUEST"
  #define WIFI_PW            "UserPassword"

  // Authentication/Registration Details  		
  #define DWOPEN_APPTOKEN    "641k5k4l5k6l76lg"      //Application Token
  #define DWOPEN_SERVER      "api.devicewise.com"    //deviceWISE server
  #define DWOPEN_PORT        1883                    //MQTT Port

  //ILST Headquarters - GPS Detail...
  #define ILST_Latitude      "26.394919"
  #define ILST_Longitude     "-80.112808"
  #define ILST_Altitude      "3.0"
  #define ILST_Speed         "0.0"
  #define ILST_Heading       "0.0"

  char dwThingKey[40];
  byte mac[6] = {0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5};
  
  void callback(char* topic, uint8_t* payload, unsigned int length);

  LWiFiClient c;

  #define TOUCH_PIN       2     //Touch Sensor on D4
  #define DHTPIN          4     //Temp/Hum Sensor on D2

  #define INTERRUPT_PIN2  0     //Int Pin2 = 0
  
  #define DHTTYPE DHT22         //DHT 22 (AM2302)
  
  DHT dht(DHTPIN, DHTTYPE);     //Init DHT sensor

  SeeedLedBar bar(9, 8);        //Grove D8, Pins: Clk, Data

  #define GREEN_LED   9
  #define RED_LED     0

  #define ALARM_NORMAL        8
  #define ALARM_WARNING       9
  #define ALARM_OVERHEATING   10

  #define ALARM_NORMAL_STATE          0
  #define ALARM_WARNING_STATE         1
  #define ALARM_OVERHEATING_STATE     2
  
  #define ALARM_NORMAL_MSG       "Temp Normal"
  #define ALARM_WARNING_MSG      "Temp Warning"
  #define ALARM_OVERHEATING_MSG  "Temp Overheating"
  
  int TempLevel             = 0;
  int AlarmLevel            = ALARM_NORMAL_STATE;
  const char *AlarmLevelMsg = "";
  boolean AlarmLevelChange  = FALSE;

  
  dwOpenClient dwClient((char*) DWOPEN_SERVER, DWOPEN_PORT, callback, c);


//Touch Sensor ISR
void TouchISR()
{ 
    TempLevel++;    
    TempLevel = TempLevel > 10 ? 0 : TempLevel;  

    bar.setLevelReverse(TempLevel);         

    if ( TempLevel <= ALARM_NORMAL )
    {
        AlarmLevel = ALARM_NORMAL_STATE;
        AlarmLevelMsg = ALARM_NORMAL_MSG;
    }
    else if ( TempLevel == ALARM_WARNING )
    {
        AlarmLevel = ALARM_WARNING_STATE;
        AlarmLevelMsg = ALARM_WARNING_MSG;        
    }
    else
    {
        AlarmLevel = ALARM_OVERHEATING_STATE;
        AlarmLevelMsg = ALARM_OVERHEATING_MSG;      
    }
    
    AlarmLevelChange = TRUE;
}

void setup() 
{
  int i;
  
  Serial.begin(9600);
  
  for(i = 5; i > 0; i-- )
  {
    Serial.print("Ready to Execute... Countdown = ");
    Serial.println(i); 
    delay( 1000 );
  }
 
  Serial.print("Attempting to connect to Network named: "); Serial.println(WIFI_SSID); 

  LWiFi.begin();
  delay(1000);
  
  // keep retrying until connected to AP
  Serial.println("Connecting to AP");
  while (0 == LWiFi.connectWPA(WIFI_SSID, WIFI_PW))
  {
    Serial.print("WiFi Status = "); Serial.println(LWiFi.status());    
    delay(1000);
  }


  Serial.println("\nYou're connected to the network\n\rWaiting for an ip address");
  
  while (LWiFi.localIP() == INADDR_NONE) 
  {
    Serial.print(".");
    delay(300);
  }

  Serial.print("\nIP Address obtained : "); Serial.println(LWiFi.localIP());
  
  //Get my MAC address
  LWiFi.macAddress(mac);  

  dht.begin();                         //Initialize temperature humidity sensor
  
  bar.begin(9, 8);
  bar.setLevelReverse(TempLevel);   
  
  attachInterrupt(INTERRUPT_PIN2,TouchISR,FALLING);      

  //Format the ThingKey for this run using our MAC
  sprintf(dwThingKey, "LinkIt-%x%x%x%x%x%x", mac[5], mac[4], mac[3], mac[2], mac[1], mac[0]);  
  Serial.print("My Thing Key: ");  Serial.println( dwThingKey );   

  
}

void loop() 
{
  int i;
  int rc = -1;
  char szText[128];
  char szPropValue[16];

  //Are we already connected to the cloud?
  if( !dwClient.connected() )
  {
      //No, Proceed with cloud connect sequence...
      Serial.print("Connecting to: "); Serial.print(DWOPEN_SERVER); Serial.print(" with client id: "); Serial.println(dwThingKey);
  
      dwClient.connect((char*) dwThingKey, (char*) dwThingKey, (char*) DWOPEN_APPTOKEN);
  
      Serial.println("Connected to m2mAIR Cloud...");
       
      //Serial.println("Subscribing to Topic 'reply'...");
      dwClient.subscribe(MQTT_REPLY);
      
      //Serial.println("Subscribing to Topic 'notify'...");
      dwClient.subscribe(MQTT_NOTIFY);
      
      // Publish Log Message... Connected
      Serial.println("LinkIt - Log Publish In Process..." );
      rc = dwClient.dwLogPublish( (char*) dwThingKey, (char*) "LinkIt Demo Program Started...");
       
      Serial.print("Publish Log Message return code : "); Serial.println(rc);   
      
      // Publish Location Message...
      Serial.println("dwOpen API - Location Publish In Process..." );      
      rc = dwClient.dwLocationPublish( dwThingKey, ILST_Latitude, ILST_Longitude, ILST_Altitude, ILST_Speed, ILST_Heading );    
      
      Serial.print("Publish Location return code : "); Serial.println(rc);          
      
      // Publish Attribute...
      Serial.println("dwOpen API - Attribute Publish In Process..." );    
      sprintf(szPropValue, "%x%x%x%x%x%x", mac[5], mac[4], mac[3], mac[2], mac[1], mac[0]);  	  
      rc = dwClient.dwAttributePublish( dwThingKey, "MAC", szPropValue );    
      
      Serial.print("Publish Location return code : "); Serial.println(rc);         
  }

   //Are we connected now?
   if( dwClient.connected() )   
   {           
      float Temperature = 0.0;
      float Humidity = 0.0;

      //Temp & Humidity data is ready?
      if ( dht.readHT(&Temperature, &Humidity) )
      {
          //Yes, format it and publish it...
          Serial.print(dwThingKey);      
          Serial.print(", Temperature:");      
          Serial.println(Temperature); 
          sprintf(szPropValue,"%f", Temperature);
          
          dwClient.dwPropertyPublish( dwThingKey, "Temperature", szPropValue );         
          
          Serial.print(dwThingKey);      
          Serial.print(", Humidity:");      
          Serial.println(Humidity); 
          sprintf(szPropValue,"%f", Humidity);
          
          dwClient.dwPropertyPublish( dwThingKey, "Humidity", szPropValue ); 
      }  

      //Has the Alarm Level Changed (led status bar changed)?
      if ( AlarmLevelChange )
      {
          //Yes, publish new alarm details...
          AlarmLevelChange = FALSE;
                      
          Serial.print("Alarm Level Change, AlarmLevel:"); Serial.print(AlarmLevel); Serial.print(", Msg:"); Serial.println(AlarmLevelMsg);
          dwClient.dwAlarmPublish( dwThingKey, "TempStatus", AlarmLevel, AlarmLevelMsg );
      }              
  }
  else
  {
      Serial.println("Error, Connecting to m2mAIR Cloud...");
  }
     
  Serial.println("Sleeping for 10 Seconds..." );  
  for (i = 0; i < 10; i++ )
  {
      //Check if new Alarm Change was reported - if so, break out of loop
      if (AlarmLevelChange) break;
      dwClient.dwSleep(1000);
  }
}


void ExtractKeyValue( const char *haystack, const char *Key, char *ResultValue )
{
  char *KeyPtr;
  char *KeyValue;
  char *KeyValueEnd;
  int   KeyValueLen;
	
  char needleTmp[40];
    	
  strcpy( needleTmp, "\"");
  strcat( needleTmp, Key);
  strcat( needleTmp, "\":\"");

  KeyPtr = strstr(haystack, needleTmp );
  	
  if( KeyPtr != NULL) 
  {	  
      KeyValue = KeyPtr + strlen(needleTmp);
      KeyValueEnd = KeyValue;
	
    	while( *KeyValueEnd != '\"')
    	{
    	    KeyValueEnd++;
    	}
	
      KeyValueLen = KeyValueEnd - KeyValue;
      
    	strncpy( ResultValue, KeyValue, KeyValueLen );
    	*(ResultValue + KeyValueLen ) = '\0';
   }
   else
   {
    	strcpy( ResultValue, "" );
     }
}
        
// Sets the LED accordingly
void setLED( int LEDPin, char *LEDStatus )
{
    TempLevel = 0;
    bar.setLevel(TempLevel);    //Turn OFF all LEDs by default
    
    if ( !strcmp(LEDStatus, "1") )
    {        
        Serial.println("LED On!!!");                  
        bar.singleLed(LEDPin, HIGH);
    }
    else
    {
        Serial.println("LED Off!!!");          
        bar.singleLed(LEDPin, LOW);          
    }  
}

//TR50 JSON Sample Request/Response
#define MAILBOX_IN_PROGRESS_ACK  "{\"999-999\": { \"command\": \"mailbox.check\", \"params\": { \"autoComplete\": true, \"limit\": 0 }}}"
#define MAILBOX_REQ_RECEIVED     "{\"0000123\": { \"command\": \"mailbox.update\", \"params\": {\"id\": \"%s\", \"msg\": \"Executing...\"}}}"
#define MAILBOX_REQ_ACK          "{\"0000124\": { \"command\": \"mailbox.ack\", \"params\": {\"id\": \"%s\", \"errorCode\": \"0\", \"errorMessage\": \"OK\"}}}"

// The function is called when an MQTT message arrives
void callback(char *topic, uint8_t *payload, unsigned int length) 
{
  int i;
  char jsonText[1024];
  char LEDStatus[10];   
  char msgID[50];

  //Serial.println("Callback - MQTT Message Arrived...");
  //Serial.print("Topic: ");  Serial.println( topic ); 
  
  //Serial.print("Payload Length: ");  Serial.println( length ); 
  //Serial.print("Payload: ");  Serial.println( (char *) payload );
  
  if(strstr(topic, MQTT_NOTIFY) != NULL) 
  {  
      //Serial.println("Notification - Mailbox Activity... Check Mailbox");
      dwClient.publish("api", MAILBOX_IN_PROGRESS_ACK );    
  }
  else if( (strstr(topic, MQTT_REPLY) != NULL) )
  {
      if ((strstr((char *)payload, "999-999") != NULL)) 
      {     
          //Serial.println("Reply - Mailbox Check Response... Mailbox Request Received");   
          ExtractKeyValue( (char *) payload, "id", msgID ); 
          
          sprintf( jsonText, MAILBOX_REQ_RECEIVED, msgID );  
          dwClient.publish("api", jsonText );
  
          //Set LED accordingly...
                           
          ExtractKeyValue( (char *) payload, "greenled", LEDStatus );                
          if ( strlen(LEDStatus) ) setLED(GREEN_LED, LEDStatus);
    
          ExtractKeyValue( (char *) payload, "redled", LEDStatus );                
          if ( strlen(LEDStatus) ) setLED(RED_LED, LEDStatus);
            
          //Serial.println("Processing Complete - Send Mailbox Acknowledge...");
          sprintf( jsonText, MAILBOX_REQ_ACK, msgID );   
          dwClient.publish( "api", jsonText );  
      } 
    
  }
  
}



