/*===============================================================================
Title: dwOpenClientLinkIt.h
Description: deviceWISE LinkIt WiFi Support

Copyright (c)2014, 2015 ILS Technology.
http://www.ilstechnology.com
http://www.devicewise.com


Legal Disclaimer:

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

    Neither the name of ILS Technology nor Telit nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===================================================================================*/


#ifndef dwOpenClient_h
#define dwOpenClient_h

#include <Arduino.h>
#include "Client.h"

// MQTT_MAX_PACKET_SIZE : Maximum packet size
#define MQTT_MAX_PACKET_SIZE 1024

// MQTT_KEEPALIVE : keepAlive interval in Seconds
#define MQTT_KEEPALIVE 15

#define MQTTPROTOCOLVERSION 3
#define MQTTCONNECT     1 << 4  // Client request to connect to Server
#define MQTTCONNACK     2 << 4  // Connect Acknowledgment
#define MQTTPUBLISH     3 << 4  // Publish message
#define MQTTPUBACK      4 << 4  // Publish Acknowledgment
#define MQTTPUBREC      5 << 4  // Publish Received (assured delivery part 1)
#define MQTTPUBREL      6 << 4  // Publish Release (assured delivery part 2)
#define MQTTPUBCOMP     7 << 4  // Publish Complete (assured delivery part 3)
#define MQTTSUBSCRIBE   8 << 4  // Client Subscribe request
#define MQTTSUBACK      9 << 4  // Subscribe Acknowledgment
#define MQTTUNSUBSCRIBE 10 << 4 // Client Unsubscribe request
#define MQTTUNSUBACK    11 << 4 // Unsubscribe Acknowledgment
#define MQTTPINGREQ     12 << 4 // PING Request
#define MQTTPINGRESP    13 << 4 // PING Response
#define MQTTDISCONNECT  14 << 4 // Client is Disconnecting
#define MQTTReserved    15 << 4 // Reserved

#define MQTTQOS0        (0 << 1)
#define MQTTQOS1        (1 << 1)
#define MQTTQOS2        (2 << 1)

#define MQTT_REPLY		"reply"
#define MQTT_NOTIFY 	"notify"

class dwOpenClient {
private:
   Client* _client;
   uint8_t buffer[MQTT_MAX_PACKET_SIZE];
   uint16_t nextMsgId;
   unsigned long lastOutActivity;
   unsigned long lastInActivity;
   bool pingOutstanding;
   void (*callback)(char*,uint8_t*,unsigned int);
   uint16_t readPacket(uint16_t*);
   uint8_t readByte();
   boolean write(uint8_t header, uint8_t* buf, uint16_t length);
   uint16_t writeString(const char *string, uint8_t* buf, uint16_t pos);
   uint8_t *ip;
   char* domain;
   uint16_t port;
   int msgId;
   int rc;
   char szBuffer[1024];
public:
   	dwOpenClient();
   	dwOpenClient(uint8_t *, uint16_t, void(*)(char*,uint8_t*,unsigned int),Client& client);
   	dwOpenClient(char*, uint16_t, void(*)(char*,uint8_t*,unsigned int),Client& client);
   	boolean connect(char *);
   	boolean connect(char *, char *, char *);
   	boolean connect(char *, char *, uint8_t, uint8_t, char *);
   	boolean connect(char *, char *, char *, char *, uint8_t, uint8_t, char*);
   	void disconnect();
   	boolean publish(const char *, const char *);
   	boolean publish(const char *, const char *, unsigned int);
   	boolean publish(const char *, const char *, unsigned int, boolean);
   	boolean publish_P(char *, uint8_t PROGMEM *, unsigned int, boolean);
   	boolean subscribe(const char *);
   	boolean unsubscribe(const char *);
   	boolean loop();
   	boolean connected();
	void dwSleep(int);

	void jsonCreateBindThing( char *thingKey, char *jsonText );
	void jsonCreateUnbindThing( char *thingKey, char *jsonText );
	void jsonCreateLogPublish( char *thingKey, const char *msgText, char *jsonText );
	void jsonCreatePropertyPublish( char *thingKey, const char *propKey, const char *propValue, char *jsonText );
	void jsonCreatePropertyCurrent( char *thingKey, const char *propKey, char *jsonText );
	void jsonCreateAlarmPublish( char *thingKey, const char *alarmKey, int alarmState, const char *alarmMsg, char *jsonText );
	void jsonCreateLocationPublish( char *thingKey, const char *locLatitude, const char *locLongitude, const char *locAltitude, const char *locSpeed, const char *locHeading, char *jsonText );
	void jsonCreateSetAttribute( char *thingKey, const char *attribKey, const char *attribValue, char *jsonText );
	void jsonCreateGetAttribute( char *thingKey, const char *attribKey, char *jsonText );

	int dwLogPublish( char *thingKey, const char *szLogMsg );
	int dwLocationPublish( char *thingKey, const char *szLat, const char *szLong, const char *szAlt, const char *szSpeed, const char *szHeading );
    int dwPropertyPublish( char *thingKey, const char *szKey, const char *szValue );
    int dwAttributePublish( char *thingKey, const char *szKey, const char *szValue );	
    int dwAlarmPublish( char *thingKey, const char *szKey, int iState, const char *szMessage );
};

#endif
